!-----------------------------------------------------------------------
subroutine recov(uu,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
!-----------------------------------------------------------------------
!     Calculation of primitive variables from conserved variables
!     
  use pram, only : imax, jmax, kmax, nv, ns, iwvec, ieos
  implicit none
!
  integer :: nm0, is1, ie1, js1, je1, ks1, ke1
  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)
     
  real(8) :: x1(imax),x2(jmax),x3(kmax)
  real(8) :: dt
!
  if( iwvec.eq.1 ) then
    call recov3a(uu,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
  elseif( iwvec.eq.2 ) then 
    call recov4a(uu,uri,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
  elseif( iwvec.eq.3 ) then
    call recov5a(uu,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
  elseif( iwvec.eq.4 ) then
    call recov6a(uu,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
  endif

  
!
  return
end subroutine recov
!
!-----------------------------------------------------------------------
subroutine recovb(uu,uri,x1,x2,x3,is1,ie1,js1,je1,ks1,ke1)
!-----------------------------------------------------------------------
!     Calculation of primitive variables from conserved variables
!     
  use pram, only : imax, jmax, kmax, nv, ns, iwvec
  implicit none
!
  integer :: i,j,k,is1, ie1, js1, je1, ks1, ke1
  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)
     
  real(8) :: x1(imax),x2(jmax),x3(kmax)
!
  if( iwvec.eq.1 ) then
    call recov3b(uu,uri,x1,x2,x3,is1,ie1,js1,je1,ks1,ke1)
  elseif( iwvec.eq. 2 ) then 
    call recov4b(uu,uri,x1,x2,x3,is1,ie1,js1,je1,ks1,ke1)
  endif
!
  return
end subroutine recovb
!
!--------------------------------------------------------------------
subroutine calconv(uu,uri,x1,nm0,is1,ie1,js1,je1,ks1,ke1, &
                   myranki,myrankj,myrankk)
!--------------------------------------------------------------------
!
! Calculate conserved variables at the interpolated boundary region
!
  use pram, only : imax, jmax, kmax, nv, ns, iprocs, jprocs, kprocs, &
                   gam, c0, ieos, iflag
  implicit none

  integer :: i, j, k, n, nm0, n2
  integer :: is1, ie1, js1, je1, ks1, ke1, myranki, myrankj, myrankk 
  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: x1(imax)

  real(8) :: dei1, vxi1, vyi1, vzi1, pri1, bxi1, byi1, bzi1, &
             exi1, eyi1, ezi1, chi1, phi1, psi1, &
             vti1, bti1, eti1, gfli1, rohi1, roei1, denri1, eemi1, & 
             dei2, vxi2, vyi2, vzi2, pri2, bxi2, byi2, bzi2, &
             exi2, eyi2, ezi2, chi2, phi2, psi2, &
             vti2, bti2, eti2, gfli2, rohi2, roei2, denri2, eemi2, &
             dej1, vxj1, vyj1, vzj1, prj1, bxj1, byj1, bzj1, &
             exj1, eyj1, ezj1, chj1, phj1, psj1, &
             vtj1, btj1, etj1, gflj1, rohj1, roej1, denrj1, eemj1, & 
             dej2, vxj2, vyj2, vzj2, prj2, bxj2, byj2, bzj2, &
             exj2, eyj2, ezj2, chj2, phj2, psj2, &
             vtj2, btj2, etj2, gflj2, rohj2, roej2, denrj2, eemj2, &
             dek1, vxk1, vyk1, vzk1, prk1, bxk1, byk1, bzk1, &
             exk1, eyk1, ezk1, chk1, phk1, psk1, &
             vtk1, btk1, etk1, gflk1, rohk1, roek1, denrk1, eemk1, & 
             dek2, vxk2, vyk2, vzk2, prk2, bxk2, byk2, bzk2, &
             exk2, eyk2, ezk2, chk2, phk2, psk2, &
             vtk2, btk2, etk2, gflk2, rohk2, roek2, denrk2, eemk2  
!
!!!!
  do k=ks1+nm0,ke1-nm0
    do i=is1+nm0,ie1-nm0
      do n=1,nm0

        if(myrankj .eq. 0 .or. jprocs .eq. 1) then

          dej1=uri(1,i,n,k)
          vxj1=uri(2,i,n,k)
          vyj1=uri(3,i,n,k)
          vzj1=uri(4,i,n,k)
          prj1=uri(5,i,n,k)
          bxj1=uri(7,i,n,k)
          byj1=uri(8,i,n,k)
          bzj1=uri(9,i,n,k)
          exj1=uri(10,i,n,k)
          eyj1=uri(11,i,n,k)
          ezj1=uri(12,i,n,k)
          chj1=uri(13,i,n,k)
          phj1=uri(14,i,n,k)
          psj1=uri(15,i,n,k)
         
          vtj1=(vxj1**2+vyj1**2+vzj1**2)/c0**2
          btj1=(bxj1**2+byj1**2+bzj1**2)
          etj1=(exj1**2+eyj1**2+ezj1**2)
          gflj1=1.0/sqrt(1.0-vtj1)

          if(ieos .eq. 0) then
            rohj1=dej1+(gam/(gam-1.0))*prj1
          elseif(ieos .eq. 1) then
            rohj1=(5./2.)*prj1+sqrt((9./4.)*prj1**2+dej1**2)
          elseif(ieos .eq. 2) then
            roej1=(3./2.)*(prj1+((3.*prj1**2) &
                  /(2.0*dej1+sqrt(2.*prj1**2+4.*dej1**2)) ))
            rohj1=dej1+roej1+prj1
          elseif(ieos .eq. 3) then
            rohj1=(12.*prj1**2+8.*prj1*dej1+2.*dej1**2)/(3.*prj1+2.*dej1)
          endif
          
          denrj1=(gflj1**2)*rohj1
          eemj1=0.5*(btj1+etj1)
!
          uu(1,i,n,k)=gflj1*dej1
          uu(2,i,n,k)=denrj1*vxj1+(eyj1*bzj1-byj1*ezj1)
          uu(3,i,n,k)=denrj1*vyj1+(ezj1*bxj1-bzj1*exj1)
          uu(4,i,n,k)=denrj1*vzj1+(exj1*byj1-bxj1*eyj1)
          uu(5,i,n,k)=denrj1-prj1+eemj1
      
          if( iflag(6).le.1 ) then
            uu(6,i,n,k)=gflj1*dej1*x1(i)
          elseif( iflag(6).le.3 ) then
            uu(6,i,n,k)=gflj1*prj1/dej1**(gam-1.0)
          elseif( iflag(6).eq.4 ) then
            uu(6,i,n,k)=gflj1*dej1*log(prj1/dej1**gam)
          endif
!
          uu(7,i,n,k)=bxj1
          uu(8,i,n,k)=byj1
          uu(9,i,n,k)=bzj1
          uu(10,i,n,k)=exj1
          uu(11,i,n,k)=eyj1
          uu(12,i,n,k)=ezj1
          uu(13,i,n,k)=chj1
          uu(14,i,n,k)=phj1
          uu(15,i,n,k)=psj1

        endif
!!!!
        if(myrankj .eq. jprocs-1 .or. jprocs .eq. 1) then
          n2=n-1

          dej2=uri(1,i,jmax-n2,k)
          vxj2=uri(2,i,jmax-n2,k)
          vyj2=uri(3,i,jmax-n2,k)
          vzj2=uri(4,i,jmax-n2,k)
          prj2=uri(5,i,jmax-n2,k)
          bxj2=uri(7,i,jmax-n2,k)
          byj2=uri(8,i,jmax-n2,k)
          bzj2=uri(9,i,jmax-n2,k)
          exj2=uri(10,i,jmax-n2,k)
          eyj2=uri(11,i,jmax-n2,k)
          ezj2=uri(12,i,jmax-n2,k)
          chj2=uri(13,i,jmax-n2,k)
          phj2=uri(14,i,jmax-n2,k)
          psj2=uri(15,i,jmax-n2,k)        

          vtj2=(vxj2**2+vyj2**2+vzj2**2)/c0**2
          btj2=(bxj2**2+byj2**2+bzj2**2)
          etj2=(exj2**2+eyj2**2+ezj2**2)
          gflj2=1.0/sqrt(1.0-vtj2)

          if(ieos .eq. 0) then
            rohj2=dej2+(gam/(gam-1.0))*prj2
          elseif(ieos .eq. 1) then
            rohj2=(5./2.)*prj2+sqrt((9./4.)*prj2**2+dej2**2)
          elseif(ieos .eq. 2) then
            roej2=(3./2.)*(prj2+((3.*prj2**2) &
                  /(2.0*dej2+sqrt(2.*prj2**2+4.*dej2**2)) ))
            rohj2=dej2+roej2+prj2
          elseif(ieos .eq. 3) then
            rohj2=(12.*prj2**2+8.*prj2*dej2+2.*dej2**2)/(3.*prj2+2.*dej2)
          endif
          
          denrj2=(gflj2**2)*rohj2
          eemj2=0.5*(btj2+etj2)
!
          uu(1,i,jmax-n2,k)=gflj2*dej2
          uu(2,i,jmax-n2,k)=denrj2*vxj2+(eyj2*bzj2-byj2*ezj2)
          uu(3,i,jmax-n2,k)=denrj2*vyj2+(ezj2*bxj2-bzj2*exj2)
          uu(4,i,jmax-n2,k)=denrj2*vzj2+(exj2*byj2-bxj2*eyj2)
          uu(5,i,jmax-n2,k)=denrj2-prj2+eemj2
!
          if( iflag(6).le.1 ) then
            uu(6,i,jmax-n2,k)=gflj2*dej2*x1(i)
          elseif( iflag(6).le.3 ) then
            uu(6,i,jmax-n2,k)=gflj2*prj2/dej2**(gam-1.0)
          elseif( iflag(6).eq.4 ) then
            uu(6,i,jmax-n2,k)=gflj2*dej2*log(prj2/dej2**gam)
          endif
!
          uu(7,i,jmax-n2,k)=bxj2
          uu(8,i,jmax-n2,k)=byj2
          uu(9,i,jmax-n2,k)=bzj2
          uu(10,i,jmax-n2,k)=exj2
          uu(11,i,jmax-n2,k)=eyj2
          uu(12,i,jmax-n2,k)=ezj2
          uu(13,i,jmax-n2,k)=chj2
          uu(14,i,jmax-n2,k)=phj2
          uu(15,i,jmax-n2,k)=psj2
!
        endif
      enddo
    enddo
  enddo
!
  do k=ks1+nm0,ke1-nm0
    do j=js1,je1
      do n=1,nm0
        
        if(myranki .eq. 0 .or. iprocs .eq. 1) then
          dei1=uri(1,n,j,k)
          vxi1=uri(2,n,j,k)
          vyi1=uri(3,n,j,k)
          vzi1=uri(4,n,j,k)
          pri1=uri(5,n,j,k)
          bxi1=uri(7,n,j,k)
          byi1=uri(8,n,j,k)
          bzi1=uri(9,n,j,k)
          exi1=uri(10,n,j,k)
          eyi1=uri(11,n,j,k)
          ezi1=uri(12,n,j,k)
          chi1=uri(13,n,j,k)
          phi1=uri(14,n,j,k)
          psi1=uri(15,n,j,k)
!         
          vti1=(vxi1**2+vyi1**2+vzi1**2)/c0**2
          bti1=(bxi1**2+byi1**2+bzi1**2)
          eti1=(exi1**2+eyi1**2+ezi1**2)
          gfli1=1.0/sqrt(1.0-vti1)

          if(ieos .eq. 0) then
            rohi1=dei1+(gam/(gam-1.0))*pri1
          elseif(ieos .eq. 1) then
            rohi1=(5./2.)*pri1+sqrt((9./4.)*pri1**2+dei1**2)
          elseif(ieos .eq. 2) then
            roei1=(3./2.)*(pri1+((3.*pri1**2) &
                  /(2.0*dei1+sqrt(2.*pri1**2+4.*dei1**2)) ))
            rohi1=dei1+roei1+pri1
          elseif(ieos .eq. 3) then
            rohi1=(12.*pri1**2+8.*pri1*dei1+2.*dei1**2)/(3.*pri1+2.*dei1)
          endif
          
          denri1=(gfli1**2)*rohi1
          eemi1=0.5*(bti1+eti1)
!  
          uu(1,n,j,k)=gfli1*dei1
          uu(2,n,j,k)=denri1*vxi1+(eyi1*bzi1-byi1*ezi1)
          uu(3,n,j,k)=denri1*vyi1+(ezi1*bxi1-bzi1*exi1)
          uu(4,n,j,k)=denri1*vzi1+(exi1*byi1-bxi1*eyi1)
          uu(5,n,j,k)=denri1-pri1+eemi1
!
          if( iflag(6).le.1 ) then
            uu(6,n,j,k)=gfli1*dei1*x1(n)
          elseif( iflag(6).le.3 ) then
           uu(6,n,j,k)=gfli1*pri1/dei1**(gam-1.0)
          elseif( iflag(6).eq.4 ) then
            uu(6,n,j,k)=gfli1*dei1*log(pri1/dei1**gam)
          endif
!
          uu(7,n,j,k)=bxi1
          uu(8,n,j,k)=byi1
          uu(9,n,j,k)=bzi1
          uu(10,n,j,k)=exi1
          uu(11,n,j,k)=eyi1
          uu(12,n,j,k)=ezi1
          uu(13,n,j,k)=chi1
          uu(14,n,j,k)=phi1
          uu(15,n,j,k)=psi1

        endif
!!!!
        if(myranki .eq. iprocs-1 .or. iprocs .eq. 1) then
          n2=n-1
!
          dei2=uri(1,imax-n2,j,k)
          vxi2=uri(2,imax-n2,j,k)
          vyi2=uri(3,imax-n2,j,k)
          vzi2=uri(4,imax-n2,j,k)
          pri2=uri(5,imax-n2,j,k)
          bxi2=uri(7,imax-n2,j,k)
          byi2=uri(8,imax-n2,j,k)
          bzi2=uri(9,imax-n2,j,k)
          exi2=uri(10,imax-n2,j,k)
          eyi2=uri(11,imax-n2,j,k)
          ezi2=uri(12,imax-n2,j,k)
          chi2=uri(13,imax-n2,j,k)
          phi2=uri(14,imax-n2,j,k)
          psi2=uri(15,imax-n2,j,k)
         
          vti2=(vxi2**2+vyi2**2+vzi2**2)/c0**2
          bti2=(bxi2**2+byi2**2+bzi2**2)
          eti2=(exi2**2+eyi2**2+ezi2**2)
          gfli2=1.0/sqrt(1.0-vti2)

          if(ieos .eq. 0) then
            rohi2=dei2+(gam/(gam-1.0))*pri2
          elseif(ieos .eq. 1) then
            rohi2=(5./2.)*pri2+sqrt((9./4.)*pri2**2+dei2**2)
          elseif(ieos .eq. 2) then
            roei2=(3./2.)*(pri2+((3.*pri2**2) &
                  /(2.0*dei2+sqrt(2.*pri2**2+4.*dei2**2)) ))
            rohi2=dei2+roei2+pri2
          elseif(ieos .eq. 3) then
            rohi2=(12.*pri2**2+8.*pri2*dei2+2.*dei2**2)/(3.*pri2+2.*dei2)
          endif
          
          denri2=(gfli2**2)*rohi2
          eemi2=0.5*(bti2+eti2)
!
          uu(1,imax-n2,j,k)=gfli2*dei2
          uu(2,imax-n2,j,k)=denri2*vxi2+(eyi2*bzi2-byi2*ezi2)
          uu(3,imax-n2,j,k)=denri2*vyi2+(ezi2*bxi2-bzi2*exi2)
          uu(4,imax-n2,j,k)=denri2*vzi2+(exi2*byi2-bxi2*eyi2)
          uu(5,imax-n2,j,k)=denri2-pri2+eemi2
!
          if( iflag(6).le.1 ) then
            uu(6,imax-n2,j,k)=gfli2*dei2*x1(imax-n2)
          elseif( iflag(6).le.3 ) then
            uu(6,imax-n2,j,k)=gfli2*pri2/dei2**(gam-1.0)
          elseif( iflag(6).eq.4 ) then
            uu(6,imax-n2,j,k)=gfli2*dei2*log(pri2/dei2**gam)
          endif
!
          uu(7,imax-n2,j,k)=bxi2
          uu(8,imax-n2,j,k)=byi2
          uu(9,imax-n2,j,k)=bzi2
          uu(10,imax-n2,j,k)=exi2
          uu(11,imax-n2,j,k)=eyi2
          uu(12,imax-n2,j,k)=ezi2
          uu(13,imax-n2,j,k)=chi2
          uu(14,imax-n2,j,k)=phi2
          uu(15,imax-n2,j,k)=psi2

        endif
      enddo
    enddo
  enddo
!
  do j=js1,je1
    do i=is1,ie1
      do n=1,nm0

        if(myrankk .eq. 0 .or. kprocs .eq. 1) then
          dek1=uri(1,i,j,n)
          vxk1=uri(2,i,j,n)
          vyk1=uri(3,i,j,n)
          vzk1=uri(4,i,j,n)
          prk1=uri(5,i,j,n)
          bxk1=uri(7,i,j,n)
          byk1=uri(8,i,j,n)
          bzk1=uri(9,i,j,n)
          exk1=uri(10,i,j,n)
          eyk1=uri(11,i,j,n)
          ezk1=uri(12,i,j,n)
          chk1=uri(13,i,j,n)
          phk1=uri(14,i,j,n)
          psk1=uri(15,i,j,n)
         
          vtk1=(vxk1**2+vyk1**2+vzk1**2)/c0**2
          btk1=(bxk1**2+byk1**2+bzk1**2)
          etk1=(exk1**2+eyk1**2+ezk1**2)
          gflk1=1.0/sqrt(1.0-vtk1)

          if(ieos .eq. 0) then
            rohk1=dek1+(gam/(gam-1.0))*prk1
          elseif(ieos .eq. 1) then
            rohk1=(5./2.)*prk1+sqrt((9./4.)*prk1**2+dek1**2)
          elseif(ieos .eq. 2) then
            roek1=(3./2.)*(prk1+((3.*prk1**2) &
                  /(2.0*dek1+sqrt(2.*prk1**2+4.*dek1**2)) ))
            rohk1=dek1+roek1+prk1
          elseif(ieos .eq. 3) then
            rohk1=(12.*prk1**2+8.*prk1*dek1+2.*dek1**2)/(3.*prk1+2.*dek1)
          endif
          
          denrk1=(gflk1**2)*rohk1
          eemk1=0.5*(btk1+etk1)
!
          uu(1,i,j,n)=gflk1*dek1
          uu(2,i,j,n)=denrk1*vxk1+(eyk1*bzk1-byk1*ezk1)
          uu(3,i,j,n)=denrk1*vyk1+(ezk1*bxk1-bzk1*exk1)
          uu(4,i,j,n)=denrk1*vzk1+(exk1*byk1-bxk1*eyk1)
          uu(5,i,j,n)=denrk1-prk1+eemk1
!
          if( iflag(6).le.1 ) then
            uu(6,i,j,n)=gflk1*dek1*x1(i)
          elseif( iflag(6).le.3 ) then
            uu(6,i,j,n)=gflk1*prk1/dek1**(gam-1.0)
          elseif( iflag(6).eq.4 ) then
            uu(6,i,j,n)=gflk1*dek1*log(prk1/dek1**gam)
          endif
!
          uu(7,i,j,n)=bxk1
          uu(8,i,j,n)=byk1
          uu(9,i,j,n)=bzk1
          uu(10,i,j,n)=exk1
          uu(11,i,j,n)=eyk1
          uu(12,i,j,n)=ezk1
          uu(13,i,j,n)=chk1
          uu(14,i,j,n)=phk1
          uu(15,i,j,n)=psk1
        endif
!!!!
        if(myrankk .eq. kprocs-1 .or. kprocs .eq. 1) then
          n2=n-1
 
          dek2=uri(1,i,j,kmax-n2)
          vxk2=uri(2,i,j,kmax-n2)
          vyk2=uri(3,i,j,kmax-n2)
          vzk2=uri(4,i,j,kmax-n2)
          prk2=uri(5,i,j,kmax-n2)
          bxk2=uri(7,i,j,kmax-n2)
          byk2=uri(8,i,j,kmax-n2)
          bzk2=uri(9,i,j,kmax-n2)
          exk2=uri(10,i,j,kmax-n2)
          eyk2=uri(11,i,j,kmax-n2)
          ezk2=uri(12,i,j,kmax-n2)
          chk2=uri(13,i,j,kmax-n2)
          phk2=uri(14,i,j,kmax-n2)
          psk2=uri(15,i,j,kmax-n2)
!
          vtk2=(vxk2**2+vyk2**2+vzk2**2)/c0**2
          btk2=(bxk2**2+byk2**2+bzk2**2)
          etk2=(exk2**2+eyk2**2+ezk2**2)
          gflk2=1.0/sqrt(1.0-vtk2)

          if(ieos .eq. 0) then
            rohk2=dek2+(gam/(gam-1.0))*prk2
          elseif(ieos .eq. 1) then
           rohk2=(5./2.)*prk2+sqrt((9./4.)*prk2**2+dek2**2)
          elseif(ieos .eq. 2) then
            roek2=(3./2.)*(prk2+((3.*prk2**2) &
                  /(2.0*dek2+sqrt(2.*prk2**2+4.*dek2**2)) ))
            rohk2=dek2+roek2+prk2
          elseif(ieos .eq. 3) then
            rohk2=(12.*prk2**2+8.*prk2*dek2+2.*dek2**2)/(3.*prk2+2.*dek2)
          endif
          
          denrk2=(gflk2**2)*rohk2
          eemk2=0.5*(btk2+etk2)
!
          uu(1,i,j,kmax-n2)=gflk2*dek2
          uu(2,i,j,kmax-n2)=denrk2*vxk2+(eyk2*bzk2-byk2*ezk2)
          uu(3,i,j,kmax-n2)=denrk2*vyk2+(ezk2*bxk2-bzk2*exk2)
          uu(4,i,j,kmax-n2)=denrk2*vzk2+(exk2*byk2-bxk2*eyk2)
          uu(5,i,j,kmax-n2)=denrk2-prk2+eemk2
!
          if( iflag(6).le.1 ) then
            uu(6,i,j,kmax-n2)=gflk2*dek2*x1(i)
          elseif( iflag(6).le.3 ) then
            uu(6,i,j,kmax-n2)=gflk2*prk2/dek2**(gam-1.0)
          elseif( iflag(6).eq.4 ) then
            uu(6,i,j,kmax-n2)=gflk2*dek2*log(prk2/dek2**gam)
          endif
!
          uu(7,i,j,kmax-n2)=bxk2
          uu(8,i,j,kmax-n2)=byk2
          uu(9,i,j,kmax-n2)=bzk2        
          uu(10,i,j,kmax-n2)=exk2
          uu(11,i,j,kmax-n2)=eyk2
          uu(12,i,j,kmax-n2)=ezk2
          uu(13,i,j,kmax-n2)=chk2
          uu(14,i,j,kmax-n2)=phk2
          uu(15,i,j,kmax-n2)=psk2          
!
        endif
      enddo
    enddo
  enddo
!
  return
end subroutine calconv
!
!---------------------------------------------------------------------@
subroutine recov3a(uu,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
!
! Recover step 
! Calculate conserved variables (uu) to primitive variables (uri)
! Pressure method (Palenzuela et al. 2009)
!
  use pram, only : imax, jmax, kmax, nv, ns, gam, c0, iter, ieos
  implicit none
!
  integer :: i, j, k, nnn, nm0, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: x1(imax), x2(jmax), x3(kmax)

  real(8) :: dro, tx, ty, tz, ee, bx, by, bz, ex, ey, ez, sg, &
             ro, vx, vy, vz, pp, ro1, vx1, vy1, vz1, pp1, sg1, &
             bsq, esq, sx, sy, sz, vsq, gf, delta, deltaend, df, f, &
             xxn, po, pn, ep, hen, te, cssq, safty, dv, &
             aa, bb, cc, tmp1, &
             e0x, e0y, e0z, eep, epx, epy, epz, etx, ety, etz, ft, fp, &
             ex1, ey1, ez1       
  real(8) :: dt
       
  integer :: irecov
!   
!--- define variable of safe number ----
      safty = 1.0d-10

      dv=1.0d-5
!=====================================================================@
!
! ----  Calculation of variables ---
!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        
        dro=uu(1,i,j,k)
        tx=uu(2,i,j,k)
        ty=uu(3,i,j,k)
        tz=uu(4,i,j,k)
        ee=uu(5,i,j,k)
        bx=uu(7,i,j,k)
        by=uu(8,i,j,k)
        bz=uu(9,i,j,k)
        ex=uu(10,i,j,k)
        ey=uu(11,i,j,k)
        ez=uu(12,i,j,k)
         
        ro=uri(1,i,j,k)
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        pp=uri(5,i,j,k)
        sg=uri(16,i,j,k)

        ro1=uri(1,i,j,k)
        vx1=uri(2,i,j,k)
        vy1=uri(3,i,j,k)
        vz1=uri(4,i,j,k)
        pp1=uri(5,i,j,k)
        sg1=uri(16,i,j,k)
        
        bsq=(bx*bx+by*by+bz*bz)
        esq=(ex*ex+ey*ey+ez*ez)
        sx=ey*bz-by*ez
        sy=ez*bx-bz*ex
        sz=ex*by-bx*ey
        vsq=(vx*vx+vy*vy+vz*vz)
        gf=1.0/sqrt(1.0-vsq)

!--- set parameter of delta and deltaend ----

        delta=1.d0
        deltaend = 0.00001d0
        df=1.d0
        f=1.d0
        irecov=0

!--- set initial guess ---
        xxn=pp
        po=pp

!--- Newton-Raphson routine ----
         
        do nnn=1,iter

          if(delta .ge. deltaend .and. irecov .eq. 0) then
           
            if(nnn .eq. iter) then
              irecov=1
              write(6,*) ' >> Not convergence in recov3' &
                        ,' at i, j, k:',i, j, k
              write(6,*) ' >> delta =', delta &
                        ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
            else

              e0x=vz*by-vy*bz
              e0y=vx*bz-vz*bx
              e0z=vy*bx-vx*by

              eep=ex*vx+ey*vy+ez*vz

              epx=eep*vx
              epy=eep*vy
              epz=eep*vz

              etx=ex-epx
              ety=ey-epy
              etz=ez-epz

              ft=exp(-0.5d0*sg*gf*dt)
              fp=exp(-0.5d0*sg*dt/gf)

              ex1=e0x+(etx-e0x)*ft+epx*fp
              ey1=e0y+(ety-e0y)*ft+epy*fp
              ez1=e0z+(etz-e0z)*ft+epz*fp

              esq=(ex1*ex1+ey1*ey1+ez1*ez1)
              sx=ey1*bz-by*ez1
              sy=ez1*bx-bz*ex1
              sz=ex1*by-bx*ey1
   
              vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
              vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
              vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)
              vsq=vx**2+vy**2+vz**2
              if(vsq .gt. 1.d0) then
!                write(6,*) 'vt > 1.0' 
                vsq=1.0-dv
              endif

              gf=1./sqrt(1.-vsq)

              ro=dro/gf
              ep=(ee-0.5*(esq+bsq)-dro*gf+pp*(1.-gf**2))/(dro*gf)              
!
! --- calculation of p(ro,ep) and cs from EoS
!
              if(ieos .eq. 0) then
                pn=(gam-1.0)*ro*ep
                cssq=(gam*(gam-1.)*ep)/(1.+gam*ep)
              elseif(ieos .eq. 1) then
                pn=((ep+2.)/(ep+1.))*(ro*ep/3.)
                te=ro/pn
                hen=(5./2.)*te+sqrt((9./4.)*te**2+1.)
                cssq=(te/(3.*hen))*((5.*hen-8.*te)/(hen-te))
               elseif(ieos .eq. 3) then
                 aa=9.d0
                 bb=3.*(ro-ro*ep)
                 cc=-2.*ro*(ro*ep-ro)
                 tmp1=bb**2-4.*aa*cc
!                 if(tmp1 .le. 0.d0) then
!                   write(*,*) 'error: b^2-4ac < 0'
!                   tmp1=abs(tmp1)
!                 endif
!                 pn=(-bb+sqrt(tmp1))/(2.*aa)
                 pn=(3.*pp+2.*ro)*(ro*ep)/(9.*pp+3.*ro)
                 te=ro/pn
                 hen=((12.*te**2)+8.*te+2)/(3.*te+2.)
                 cssq=te*(3.*te+2.)*(18.*te**2+24.*te+5.)/ &
                      (3.*(6.*te**2+4.*te+1.)*(9.*te**2+12.*te+2.))
               endif

              f=pn-pp
              df=vsq*cssq-1.

              delta=f/df

              po=pp
              xxn=pp-delta              

              if(xxn .lt. 0.d0) then
                irecov=2
                write(6,*) ' >> pp  < 0 in recov3' &
                          ,' at i, j, k:',i, j, k
                write(6,*) ' >> pp', xxn &
                          ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
              endif

              pp=xxn
   
              vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
              vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
              vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)
              vsq=vx**2+vy**2+vz**2
              if(vsq .gt. 1.d0) then
!                write(6,*) 'vt > 1.0' 
                vsq=1.0-dv
              endif

              gf=1./sqrt(1.-vsq)

            endif
          endif

        enddo

!--- After Newton-Rapson routine ---
! 
        if(irecov .eq. 0) then
          
          pp=xxn
   
          vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
          vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
          vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)
          vsq=vx**2+vy**2+vz**2

          if(vsq .gt. 1.d0) then
!              write(6,*) 'vt > 1.0' 
            vsq=1.0-dv
          endif

          gf=1./sqrt(1.-vsq)          
          ro=dro/gf
        endif

!--- Set Primitive variables ---

        if(irecov .eq. 0) then

          uri(1,i,j,k)=ro
          uri(2,i,j,k)=vx
          uri(3,i,j,k)=vy
          uri(4,i,j,k)=vz
          uri(5,i,j,k)=pp
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)
          uri(16,i,j,k)=sg
        else

          uri(1,i,j,k)=ro1
          uri(2,i,j,k)=vx1
          uri(3,i,j,k)=vy1
          uri(4,i,j,k)=vz1
          uri(5,i,j,k)=pp1
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)          
          uri(16,i,j,k)=sg1
        endif

      enddo
    enddo
  enddo
!
!=====================================================================@
!
  return
end subroutine recov3a
!
!---------------------------------------------------------------------@
subroutine recov3b(uu,uri,x1,x2,x3,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
! Recover step 
! Calculate conserved variables (uu) to primitive variables (uri)
! Pressure method (Palenzuela et al. 2009)
!
  use pram, only : imax, jmax, kmax, nv, ns, gam, c0, iter, ieos
  implicit none
!
  integer :: i, j, k, nnn, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: x1(imax), x2(jmax), x3(kmax)

  real(8) :: dro, tx, ty, tz, ee, bx, by, bz, ex, ey, ez, sg, &
             ro, vx, vy, vz, pp, ro1, vx1, vy1, vz1, pp1, sg1, &
             bsq, esq, sx, sy, sz, vsq, gf, delta, deltaend, df, f, &
             xxn, po, pn, ep, hen, te, cssq, safty, dv       

  integer :: irecov
!   
!--- define variable of safe number ----
      safty = 1.0d-10

      dv=1.0d-5
!=====================================================================@
!
! ----  Calculation of variables ---
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
        
        dro=uu(1,i,j,k)
        tx=uu(2,i,j,k)
        ty=uu(3,i,j,k)
        tz=uu(4,i,j,k)
        ee=uu(5,i,j,k)
        bx=uu(7,i,j,k)
        by=uu(8,i,j,k)
        bz=uu(9,i,j,k)
        ex=uu(10,i,j,k)
        ey=uu(11,i,j,k)
        ez=uu(12,i,j,k)
        sg=uri(16,i,j,k)
         
        ro=uri(1,i,j,k)
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        pp=uri(5,i,j,k)

        ro1=uri(1,i,j,k)
        vx1=uri(2,i,j,k)
        vy1=uri(3,i,j,k)
        vz1=uri(4,i,j,k)
        pp1=uri(5,i,j,k)
        sg1=uri(16,i,j,k)
        
        bsq=(bx*bx+by*by+bz*bz)
        esq=(ex*ex+ey*ey+ez*ez)
        sx=ey*bz-by*ez
        sy=ez*bx-bz*ex
        sz=ex*by-bx*ey

!--- set parameter of delta and deltaend ----

        delta=1.d0
        deltaend = 0.00001d0
        df=1.d0
        f=1.d0
        irecov=0

!--- set initial guess ---
        xxn=pp
        po=pp

!--- Newton-Raphson routine ----
         
        do nnn=1,iter

          if(delta .ge. deltaend .and. irecov .eq. 0) then
           
            if(nnn .eq. iter) then
              irecov=1
              write(6,*) ' >> Not convergence in recov3' &
                        ,' at i, j, k:',i, j, k
              write(6,*) ' >> delta =', delta &
                        ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
            else

              pp=xxn
   
              vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
              vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
              vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)
              vsq=vx**2+vy**2+vz**2
              if(vsq .gt. 1.d0) then
!                write(6,*) 'vt > 1.0' 
                vsq=1.0-dv
              endif

              gf=1./sqrt(1.-vsq)

              ro=dro/gf
              ep=(ee-0.5*(esq+bsq)-dro*gf+pp*(1.-gf**2))/(dro*gf) 
!
! --- calculation of p(ro,ep) and cs from EoS
!
              if(ieos .eq. 0) then
                pn=(gam-1.0)*ro*ep
                cssq=(gam*(gam-1.)*ep)/(1.+gam*ep)
              elseif(ieos .eq. 1) then
                pn=((ep+2.)/(ep+1.))*(ro*ep/3.)
                te=ro/pn
                hen=(5./2.)*te+sqrt((9./4.)*te**2+1.)
                cssq=(te/(3.*hen))*((5.*hen-8.*te)/(hen-te))
              endif

              f=pn-pp
              df=vsq*cssq-1.

              delta=f/df

              po=pp
              xxn=pp-delta              

              if(xxn .lt. 0.d0) then
                irecov=2
                write(6,*) ' >> pp  < 0 in recov3' &
                          ,' at i, j, k:',i, j, k
                write(6,*) ' >> pp', xxn &
                          ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
              endif

            endif
          endif

        enddo

!--- After Newton-Rapson routine ---
! 
        if(irecov .eq. 0) then
          
          pp=xxn
   
          vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
          vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
          vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)
          vsq=vx**2+vy**2+vz**2

          if(vsq .gt. 1.d0) then
!              write(6,*) 'vt > 1.0' 
            vsq=1.0-dv
          endif

          gf=1./sqrt(1.-vsq)          
          ro=dro/gf
        endif

!--- Set Primitive variables ---

        if(irecov .eq. 0) then

          uri(1,i,j,k)=ro
          uri(2,i,j,k)=vx
          uri(3,i,j,k)=vy
          uri(4,i,j,k)=vz
          uri(5,i,j,k)=pp
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)
          uri(16,i,j,k)=sg
        else

          uri(1,i,j,k)=ro1
          uri(2,i,j,k)=vx1
          uri(3,i,j,k)=vy1
          uri(4,i,j,k)=vz1
          uri(5,i,j,k)=pp1
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)          
          uri(16,i,j,k)=sg1        
        endif

      enddo
    enddo
  enddo
!
!=====================================================================@
!
  return
end subroutine recov3b
!
!---------------------------------------------------------------------@
subroutine recov4a(uu,uri,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
! Recover step 
! Calculate conserved variables (uu) to primitive variables (uri)
! same as RMHD case (apply only gamma-law EoS)
!
  use pram, only : imax, jmax, kmax, nv, ns, gam, c0, iter
  implicit none
!
  integer :: i, j, k, nnn, nm0, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: x1(imax), x2(jmax), x3(kmax)

  real(8) :: dro, tx, ty, tz, ee, bx, by, bz, ex, ey, ez, sg, &
             ro, vx, vy, vz, pp, ro1, vx1, vy1, vz1, pp1, sg1, &
             bsq, esq, sx, sy, sz, vsq, gf, delta, deltaend, df, f, &
             xxn, gam1, c1, c2, a0, a1, a2, a3, a4, roh, safty, dv       
  integer :: irecov
!   
!--- define variable of safe number ----
      safty = 1.0d-10

      dv=1.0d-5
!=====================================================================@
!
! ----  Calculation of variables ---
!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        
        dro=uu(1,i,j,k)
        tx=uu(2,i,j,k)
        ty=uu(3,i,j,k)
        tz=uu(4,i,j,k)
        ee=uu(5,i,j,k)
        bx=uu(7,i,j,k)
        by=uu(8,i,j,k)
        bz=uu(9,i,j,k)
        ex=uu(10,i,j,k)
        ey=uu(11,i,j,k)
        ez=uu(12,i,j,k)
         
        ro=uri(1,i,j,k)
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        pp=uri(5,i,j,k)
        sg=uri(16,i,j,k)

        ro1=uri(1,i,j,k)
        vx1=uri(2,i,j,k)
        vy1=uri(3,i,j,k)
        vz1=uri(4,i,j,k)
        pp1=uri(5,i,j,k)
        sg1=uri(16,i,j,k)
        
        bsq=(bx*bx+by*by+bz*bz)
        esq=(ex*ex+ey*ey+ez*ez)
        sx=ey*bz-by*ez
        sy=ez*bx-bz*ex
        sz=ex*by-bx*ey
        vsq=(vx*vx+vy*vy+vz*vz)
        gf=1.0/sqrt(1.0-vsq)

        gam1=(gam-1.0)/gam

!--- set parameter of delta and deltaend ----

        delta=1.d0
        deltaend = 0.00001d0
        df=1.d0
        f=1.d0
        irecov=0

!--- set initial guess ---
        xxn=gf

!--- Newton-Raphson routine ----
         
        do nnn=1,iter

          if(delta .ge. deltaend .and. irecov .eq. 0) then
           
            if(nnn .eq. iter) then
              irecov=1
              write(6,*) ' >> Not convergence in recov3' &
                        ,' at i, j, k:',i, j, k
              write(6,*) ' >> delta =', delta &
                        ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
            else

              gf=xxn

              c1=(tx-sx)**2+(ty-sy)**2+(tz-sz)**2
              c2=ee-0.5*(esq+bsq)

              a0=(gam1**2)*(c1+dro**2)
              a1=-2.*c2*gam1*dro
              a2=(c2**2)-2.*c1*gam1-(gam1**2)*(dro**2)
              a3=2.*c2*gam1*dro
              a4=c1-(c2**2)

              f=a4*(gf**4)+a3*(gf**3)+a2*(gf**2)+a1*gf+a0
              df=4.*a4*(gf**3)+3.*a3*(gf**2)+2.*a2*gf+a1               

              delta=f/df

              xxn=gf-delta              

              if(xxn .lt. 0.d0) then
                irecov=2
                write(6,*) ' >> gf  < 0 in recov3' &
                          ,' at i, j, k:',i, j, k
                write(6,*) ' >> gf', xxn &
                          ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
              endif

            endif
          endif

        enddo

!--- After Newton-Rapson routine ---
! 
        if(irecov .eq. 0) then
          
          gf=xxn
   
          ro=dro/gf
          roh=(ee-0.5*(bsq+esq)-(gam1*ro))/((gf**2)-gam1)
          pp=gam1*(roh-ro)
!          vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
!          vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
!          vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)
          vx=(tx-sx)/(roh*gf**2)
          vy=(ty-sy)/(roh*gf**2)
          vz=(tz-sz)/(roh*gf**2)

          vsq=vx**2+vy**2+vz**2

          if(vsq .gt. 1.d0) then
            write(6,*) 'vt > 1.0' 
            vsq=1.0-dv
            gf=1.0/sqrt(1.0-vsq)
            ro=dro/gf
            roh=(ee-0.5*(bsq+esq)-(gam1*ro))/((gf**2)-gam1)
            pp=gam1*(roh-ro)
            vx=(tx-sx)/(roh*gf**2)
            vy=(ty-sy)/(roh*gf**2)
            vz=(tz-sz)/(roh*gf**2)
!            vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
!            vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
!            vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)            
          endif
        endif

!--- Set Primitive variables ---

        if(irecov .eq. 0) then

          uri(1,i,j,k)=ro
          uri(2,i,j,k)=vx
          uri(3,i,j,k)=vy
          uri(4,i,j,k)=vz
          uri(5,i,j,k)=pp
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)
          uri(16,i,j,k)=sg
        else

          uri(1,i,j,k)=ro1
          uri(2,i,j,k)=vx1
          uri(3,i,j,k)=vy1
          uri(4,i,j,k)=vz1
          uri(5,i,j,k)=pp1
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)          
          uri(16,i,j,k)=sg1
        endif

      enddo
    enddo
  enddo
!
!=====================================================================@
!
  return
end subroutine recov4a
!
!---------------------------------------------------------------------@
subroutine recov4b(uu,uri,x1,x2,x3,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
! Recover step 
! Calculate conserved variables (uu) to primitive variables (uri)
! same as RMHD case (apply only gamma-law EoS)
!
  use pram, only : imax, jmax, kmax, nv, ns, gam, c0, iter
  implicit none
!
  integer :: i, j, k, nnn, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: x1(imax), x2(jmax), x3(kmax)

  real(8) :: dro, tx, ty, tz, ee, bx, by, bz, ex, ey, ez, sg, &
             ro, vx, vy, vz, pp, ro1, vx1, vy1, vz1, pp1, sg1, &
             bsq, esq, sx, sy, sz, vsq, gf, delta, deltaend, df, f, &
             xxn, gam1, c1, c2, a0, a1, a2, a3, a4, roh, safty, dv       
  integer :: irecov
!   
!--- define variable of safe number ----
      safty = 1.0d-10

      dv=1.0d-5
!=====================================================================@
!
! ----  Calculation of variables ---
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
        
        dro=uu(1,i,j,k)
        tx=uu(2,i,j,k)
        ty=uu(3,i,j,k)
        tz=uu(4,i,j,k)
        ee=uu(5,i,j,k)
        bx=uu(7,i,j,k)
        by=uu(8,i,j,k)
        bz=uu(9,i,j,k)
        ex=uu(10,i,j,k)
        ey=uu(11,i,j,k)
        ez=uu(12,i,j,k)
         
        ro=uri(1,i,j,k)
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        pp=uri(5,i,j,k)
        sg=uri(16,i,j,k)

        ro1=uri(1,i,j,k)
        vx1=uri(2,i,j,k)
        vy1=uri(3,i,j,k)
        vz1=uri(4,i,j,k)
        pp1=uri(5,i,j,k)
        sg1=uri(16,i,j,k)
        
        bsq=(bx*bx+by*by+bz*bz)
        esq=(ex*ex+ey*ey+ez*ez)
        sx=ey*bz-by*ez
        sy=ez*bx-bz*ex
        sz=ex*by-bx*ey
        vsq=(vx*vx+vy*vy+vz*vz)
        gf=1.0/sqrt(1.0-vsq)

        gam1=(gam-1.0)/gam

!--- set parameter of delta and deltaend ----

        delta=1.d0
        deltaend = 0.00001d0
        df=1.d0
        f=1.d0
        irecov=0

!--- set initial guess ---
        xxn=gf

!--- Newton-Raphson routine ----
         
        do nnn=1,iter

          if(delta .ge. deltaend .and. irecov .eq. 0) then
           
            if(nnn .eq. iter) then
              irecov=1
              write(6,*) ' >> Not convergence in recov3' &
                        ,' at i, j, k:',i, j, k
              write(6,*) ' >> delta =', delta &
                        ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
            else

              gf=xxn

              c1=(tx-sx)**2+(ty-sy)**2+(tz-sz)**2
              c2=ee-0.5*(esq+bsq)

              a0=(gam1**2)*(c1+dro**2)
              a1=-2.*c2*gam1*dro
              a2=(c2**2)-2.*c1*gam1-(gam1**2)*(dro**2)
              a3=2.*c2*gam1*dro
              a4=c1-(c2**2)

              f=a4*(gf**4)+a3*(gf**3)+a2*(gf**2)+a1*gf+a0
              df=4.*a4*(gf**3)+3.*a3*(gf**2)+2.*a2*gf+a1               

              delta=f/df

              xxn=gf-delta              

              if(xxn .lt. 0.d0) then
                irecov=2
                write(6,*) ' >> gf  < 0 in recov3' &
                          ,' at i, j, k:',i, j, k
                write(6,*) ' >> gf', xxn &
                          ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
              endif

            endif
          endif

        enddo

!--- After Newton-Rapson routine ---
! 
        if(irecov .eq. 0) then
          
          gf=xxn
   
          ro=dro/gf
          roh=(ee-0.5*(bsq+esq)-(gam1*dro/gf))/((gf**2)-gam1)
          pp=gam1*(roh-ro)
!          vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
!          vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
!          vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)
          vx=(tx-sx)/(roh*gf**2)
          vy=(ty-sy)/(roh*gf**2)
          vz=(tz-sz)/(roh*gf**2)

          vsq=vx**2+vy**2+vz**2

          if(vsq .gt. 1.d0) then
!              write(6,*) 'vt > 1.0' 
            vsq=1.0-dv
            gf=1.0/sqrt(1.0-vsq)
            ro=dro/gf
            roh=(ee-0.5*(bsq+esq)-(gam1*dro/gf))/((gf**2)-gam1)
            pp=gam1*(roh-ro)
            vx=(tx-sx)/(roh*gf**2)
            vy=(ty-sy)/(roh*gf**2)
            vz=(tz-sz)/(roh*gf**2)
!            vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
!            vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
!            vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)            
          endif
        endif

!--- Set Primitive variables ---

        if(irecov .eq. 0) then

          uri(1,i,j,k)=ro
          uri(2,i,j,k)=vx
          uri(3,i,j,k)=vy
          uri(4,i,j,k)=vz
          uri(5,i,j,k)=pp
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)
          uri(16,i,j,k)=sg
        else

          uri(1,i,j,k)=ro1
          uri(2,i,j,k)=vx1
          uri(3,i,j,k)=vy1
          uri(4,i,j,k)=vz1
          uri(5,i,j,k)=pp1
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)          
          uri(16,i,j,k)=sg1
        endif

      enddo
    enddo
  enddo
!
!=====================================================================@
!
  return
end subroutine recov4b
!
!---------------------------------------------------------------------@
subroutine recov4c(uu,uri,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
  use pram, only : imax, jmax, kmax, nv, ns, gam, c0, iter
  implicit none
!
  integer :: i, j, k, nnn, nm0, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: x1(imax), x2(jmax), x3(kmax)

  real(8) :: dro, tx, ty, tz, ee, bx, by, bz, ex, ey, ez, sg, &
             ro, vx, vy, vz, pp, ro1, vx1, vy1, vz1, pp1, sg1, &
             bsq, esq, sx, sy, sz, vsq, gf, txa, tya, tza, eea, ww, vv, &
             delta, deltaend, df, f, &
             xxn, c1, c2, c3, a1, a2, a3, safty, dv       
  integer :: irecov
!   
!--- define variable of safe number ----
      safty = 1.0d-10

      dv=1.0d-5
!=====================================================================@
!
! ----  Calculation of variables ---
!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        
        dro=uu(1,i,j,k)
        tx=uu(2,i,j,k)
        ty=uu(3,i,j,k)
        tz=uu(4,i,j,k)
        ee=uu(5,i,j,k)
        bx=uu(7,i,j,k)
        by=uu(8,i,j,k)
        bz=uu(9,i,j,k)
        ex=uu(10,i,j,k)
        ey=uu(11,i,j,k)
        ez=uu(12,i,j,k)
         
        ro=uri(1,i,j,k)
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        pp=uri(5,i,j,k)
        sg=uri(16,i,j,k)

        ro1=uri(1,i,j,k)
        vx1=uri(2,i,j,k)
        vy1=uri(3,i,j,k)
        vz1=uri(4,i,j,k)
        pp1=uri(5,i,j,k)
        sg1=uri(16,i,j,k)
        
        bsq=(bx*bx+by*by+bz*bz)
        esq=(ex*ex+ey*ey+ez*ez)
        sx=ey*bz-by*ez
        sy=ez*bx-bz*ex
        sz=ex*by-bx*ey
        vsq=(vx*vx+vy*vy+vz*vz)
        gf=1.0/sqrt(1.0-vsq)

        txa=tx-sx
        tya=ty-sy
        tza=tz-sz
        eea=ee-0.5*(esq+bsq)

!--- set parameter of delta and deltaend ----

        delta=1.d0
        deltaend = 0.00001d0
        df=1.d0
        f=1.d0
        irecov=0

!--- set initial guess ---
        ww=gf**2-1.0
        xxn=ww

!--- Newton-Raphson routine ----
         
        do nnn=1,iter

          if(delta .ge. deltaend .and. irecov .eq. 0) then
           
            if(nnn .eq. iter) then
              irecov=1
              write(6,*) ' >> Not convergence in recov3' &
                        ,' at i, j, k:',i, j, k
              write(6,*) ' >> delta =', delta &
                        ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
            else

              ww=xxn

              c1=txa**2+tya**2+tza**2
              c2=eea**2

              a1=((c2+c1)*(4.*(c2+c1)-(c1+dro**2))-14.*c1*c2) &
                 /(2.*(c2-c1)**2)
              a2=((4.*(c2+c1)-(c1+dro**2))**2-57.*c1*c2)/(16.*(c2-c1)**2)
              a3=-9.*c1*c2/(16.*(c2-c1)**2)

              f=(ww**3)+a1*(ww**2)+a2*ww+a3
              df=3.*(ww**2)+2.*a1*ww+a2               

              delta=f/df

              xxn=ww-delta              

              gf=sqrt(1.+xxn)

              if(gf .lt. 0.d0) then
                irecov=2
                write(6,*) ' >> gf  < 0 in recov3' &
                          ,' at i, j, k:',i, j, k
                write(6,*) ' >> gf', xxn &
                          ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
              endif

            endif
          endif

        enddo

!--- After Newton-Rapson routine ---
! 
        if(irecov .eq. 0) then
          
          ww=xxn
   
          gf=sqrt(1.+ww)
          vv=ww/sqrt(1.+ww)
          ro=dro/gf
          vx=txa*vv/sqrt(c1)
          vy=tya*vv/sqrt(c1)
          vz=tza*vv/sqrt(c1)          
          vsq=vx**2+vy**2+vz**2
          if(vsq .gt. 1.d0) then 
            write(6,*) 'vt > 1.0' 
            vsq=1.0-dv
            gf=1.0/sqrt(1.0-vsq)
            ro=dro/gf 
            vx=txa*vv/sqrt(c1)
            vy=tya*vv/sqrt(c1)
            vz=tza*vv/sqrt(c1)   
          endif
          c3=eea-txa*vx-tya*vy-tza*vz
          pp=(c3**2-ro**2)/3.*c3

        endif

!--- Set Primitive variables ---

        if(irecov .eq. 0) then

          uri(1,i,j,k)=ro
          uri(2,i,j,k)=vx
          uri(3,i,j,k)=vy
          uri(4,i,j,k)=vz
          uri(5,i,j,k)=pp
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)
          uri(16,i,j,k)=sg
        else

          uri(1,i,j,k)=ro1
          uri(2,i,j,k)=vx1
          uri(3,i,j,k)=vy1
          uri(4,i,j,k)=vz1
          uri(5,i,j,k)=pp1
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)          
          uri(16,i,j,k)=sg1
        endif

      enddo
    enddo
  enddo
!
!=====================================================================@
!
  return
end subroutine recov4c
!
!---------------------------------------------------------------------@
subroutine recov5a(uu,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
  use pram, only : imax, jmax, kmax, nv, ns, gam, c0, iter
  implicit none
!
  integer :: i, j, k, nnn, nm0, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: x1(imax), x2(jmax), x3(kmax)

  real(8) :: dro, tx, ty, tz, ee, bx, by, bz, ex, ey, ez, sg, &
             ro, vx, vy, vz, pp, ro1, vx1, vy1, vz1, pp1, sg1, &
             bsq, esq, sx, sy, sz, vsq, gf, delta, deltaend, df, f, &
             xxn, gam1, c1, c2, a0, a1, a2, a3, a4, roh, safty, dv, &
             e0x, e0y, e0z, ep, epx, epy, epz, etx, ety, etz, ft, fp       
  real(8) :: dt
  integer :: irecov
!   
!--- define variable of safe number ----
      safty = 1.0d-10

      dv=1.0d-5
!=====================================================================@
!
! ----  Calculation of variables ---
!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        
        dro=uu(1,i,j,k)
        tx=uu(2,i,j,k)
        ty=uu(3,i,j,k)
        tz=uu(4,i,j,k)
        ee=uu(5,i,j,k)
        bx=uu(7,i,j,k)
        by=uu(8,i,j,k)
        bz=uu(9,i,j,k)
        ex=uu(10,i,j,k)
        ey=uu(11,i,j,k)
        ez=uu(12,i,j,k)
         
        ro=uri(1,i,j,k)
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        pp=uri(5,i,j,k)
        sg=uri(16,i,j,k)

        ro1=uri(1,i,j,k)
        vx1=uri(2,i,j,k)
        vy1=uri(3,i,j,k)
        vz1=uri(4,i,j,k)
        pp1=uri(5,i,j,k)
        sg1=uri(16,i,j,k)
        
        bsq=(bx*bx+by*by+bz*bz)
        esq=(ex*ex+ey*ey+ez*ez)
        sx=ey*bz-by*ez
        sy=ez*bx-bz*ex
        sz=ex*by-bx*ey
        vsq=(vx*vx+vy*vy+vz*vz)
        gf=1.0/sqrt(1.0-vsq)

        gam1=(gam-1.0)/gam

!--- set parameter of delta and deltaend ----

        delta=1.d0
        deltaend = 0.00001d0
        df=1.d0
        f=1.d0
        irecov=0

!--- set initial guess ---
        xxn=gf

!--- Newton-Raphson routine ----
         
        do nnn=1,iter

          if(delta .ge. deltaend .and. irecov .eq. 0) then
           
            if(nnn .eq. iter) then
              irecov=1
              write(6,*) ' >> Not convergence in recov3' &
                        ,' at i, j, k:',i, j, k
              write(6,*) ' >> delta =', delta &
                        ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
            else          

              e0x=vz*by-vy*bz
              e0y=vx*bz-vz*bx
              e0z=vy*bx-vx*by

              ep=ex*vx+ey*vy+ez*vz

              epx=ep*vx
              epy=ep*vy
              epz=ep*vz

              etx=ex-epx
              ety=ey-epy
              etz=ez-epz

              ft=exp(-0.5d0*sg*gf*dt)
              fp=exp(-0.5d0*sg*dt/gf)

              ex=e0x+(etx-e0x)*ft+epx*fp
              ey=e0y+(ety-e0y)*ft+epy*fp
              ez=e0z+(etz-e0z)*ft+epz*fp

              esq=(ex*ex+ey*ey+ez*ez)
              sx=ey*bz-by*ez
              sy=ez*bx-bz*ex
              sz=ex*by-bx*ey

              c1=(tx-sx)**2+(ty-sy)**2+(tz-sz)**2
              c2=ee-0.5*(esq+bsq)

              a0=(gam1**2)*(c1+dro**2)
              a1=-2.*c2*gam1*dro
              a2=(c2**2)-2.*c1*gam1-(gam1**2)*(dro**2)
              a3=2.*c2*gam1*dro
              a4=c1-(c2**2)

              f=a4*(gf**4)+a3*(gf**3)+a2*(gf**2)+a1*gf+a0
              df=4.*a4*(gf**3)+3.*a3*(gf**2)+2.*a2*gf+a1               

              delta=f/df

              xxn=gf-delta              

              if(xxn .lt. 0.d0) then
                irecov=2
                write(6,*) ' >> gf  < 0 in recov3' &
                          ,' at i, j, k:',i, j, k
                write(6,*) ' >> gf', xxn &
                          ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
              endif
              
              gf=xxn
              ro=dro/gf
              roh=(ee-0.5*(bsq+esq)-(gam1*ro))/((gf**2)-gam1)
              pp=gam1*(roh-ro)

              vx=(tx-sx)/(roh*gf**2)
              vy=(ty-sy)/(roh*gf**2)
              vz=(tz-sz)/(roh*gf**2)

            endif
          endif

        enddo

!--- After Newton-Rapson routine ---
! 
        if(irecov .eq. 0) then
          
          gf=xxn
   
          ro=dro/gf
          roh=(ee-0.5*(bsq+esq)-(gam1*ro))/((gf**2)-gam1)
          pp=gam1*(roh-ro)
!          vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
!          vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
!          vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)
          vx=(tx-sx)/(roh*gf**2)
          vy=(ty-sy)/(roh*gf**2)
          vz=(tz-sz)/(roh*gf**2)

          vsq=vx**2+vy**2+vz**2

          if(vsq .gt. 1.d0) then
            write(6,*) 'vt > 1.0' 
            vsq=1.0-dv
            gf=1.0/sqrt(1.0-vsq)
            ro=dro/gf
            roh=(ee-0.5*(bsq+esq)-(gam1*ro))/((gf**2)-gam1)
            pp=gam1*(roh-ro)
            vx=(tx-sx)/(roh*gf**2)
            vy=(ty-sy)/(roh*gf**2)
            vz=(tz-sz)/(roh*gf**2)
!            vx=(tx-sx)/(ee-0.5*(esq+bsq)+pp)
!            vy=(ty-sy)/(ee-0.5*(esq+bsq)+pp)
!            vz=(tz-sz)/(ee-0.5*(esq+bsq)+pp)            
          endif
        endif

!--- Set Primitive variables ---

        if(irecov .eq. 0) then

          uri(1,i,j,k)=ro
          uri(2,i,j,k)=vx
          uri(3,i,j,k)=vy
          uri(4,i,j,k)=vz
          uri(5,i,j,k)=pp
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=uu(7,i,j,k)
          uri(8,i,j,k)=uu(8,i,j,k)
          uri(9,i,j,k)=uu(9,i,j,k)
          uri(10,i,j,k)=uu(10,i,j,k)
          uri(11,i,j,k)=uu(11,i,j,k)
          uri(12,i,j,k)=uu(12,i,j,k)
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)
          uri(16,i,j,k)=sg
        else

          uri(1,i,j,k)=ro1
          uri(2,i,j,k)=vx1
          uri(3,i,j,k)=vy1
          uri(4,i,j,k)=vz1
          uri(5,i,j,k)=pp1
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=uu(7,i,j,k)
          uri(8,i,j,k)=uu(8,i,j,k)
          uri(9,i,j,k)=uu(9,i,j,k)
          uri(10,i,j,k)=uu(10,i,j,k)
          uri(11,i,j,k)=uu(11,i,j,k)
          uri(12,i,j,k)=uu(12,i,j,k)
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)          
          uri(16,i,j,k)=sg1
        endif

      enddo
    enddo
  enddo
!
!=====================================================================@
!
  return
end subroutine recov5a
!
!---------------------------------------------------------------------@
subroutine recov6a(uu,uri,dt,x1,x2,x3,nm0,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
  use pram, only : imax, jmax, kmax, nv, ns, gam, c0, iter, ieos
  implicit none
!
  integer :: i, j, k, nnn, nm0, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: x1(imax), x2(jmax), x3(kmax)

  real(8) :: dro, tx, ty, tz, ee, bx, by, bz, ex, ey, ez, sg, &
             ro, vx, vy, vz, pp, ro1, vx1, vy1, vz1, pp1, sg1, &
             bsq, esq, sx, sy, sz, vsq, gf, txa, tya, tza, eea, te, &
             delta, deltaend, df, f, tmp1,&
             xxn, h, dh, ssq, gfsq, dgf, safty, dv, &
             e0x, e0y, e0z, ep, epx, epy, epz, etx, ety, etz, ft, fp
  real(8) :: dt       
  integer :: irecov
!   
!--- define variable of safe number ----
      safty = 1.0d-10

      dv=1.0d-5
!=====================================================================@
!
! ----  Calculation of variables ---
!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        
        dro=uu(1,i,j,k)
        tx=uu(2,i,j,k)
        ty=uu(3,i,j,k)
        tz=uu(4,i,j,k)
        ee=uu(5,i,j,k)
        bx=uu(7,i,j,k)
        by=uu(8,i,j,k)
        bz=uu(9,i,j,k)
        ex=uu(10,i,j,k)
        ey=uu(11,i,j,k)
        ez=uu(12,i,j,k)
         
        ro=uri(1,i,j,k)
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        pp=uri(5,i,j,k)
        sg=uri(16,i,j,k)

        ro1=uri(1,i,j,k)
        vx1=uri(2,i,j,k)
        vy1=uri(3,i,j,k)
        vz1=uri(4,i,j,k)
        pp1=uri(5,i,j,k)
        sg1=uri(16,i,j,k)
        
        bsq=(bx*bx+by*by+bz*bz)
        esq=(ex*ex+ey*ey+ez*ez)
        sx=ey*bz-by*ez
        sy=ez*bx-bz*ex
        sz=ex*by-bx*ey
        vsq=(vx*vx+vy*vy+vz*vz)
        gf=1.0/sqrt(1.0-vsq)

        txa=tx-sx
        tya=ty-sy
        tza=tz-sz
        eea=ee-0.5*(esq+bsq)

!--- set parameter of delta and deltaend ----

        delta=1.d0
        deltaend = 0.00001d0
        df=1.d0
        f=1.d0
        irecov=0

!--- set initial guess ---
        te=pp*gf/dro
        xxn=te

!--- Newton-Raphson routine ----
         
        do nnn=1,iter

          if(delta .ge. deltaend .and. irecov .eq. 0) then
           
            if(nnn .eq. iter) then
              irecov=1
              write(6,*) ' >> Not convergence in recov3' &
                        ,' at i, j, k:',i, j, k
              write(6,*) ' >> delta =', delta &
                        ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
            else

              e0x=vz*by-vy*bz
              e0y=vx*bz-vz*bx
              e0z=vy*bx-vx*by

              ep=ex*vx+ey*vy+ez*vz

              epx=ep*vx
              epy=ep*vy
              epz=ep*vz

              etx=ex-epx
              ety=ey-epy
              etz=ez-epz

              ft=exp(-0.5d0*sg*gf*dt)
              fp=exp(-0.5d0*sg*dt/gf)

              ex=e0x+(etx-e0x)*ft+epx*fp
              ey=e0y+(ety-e0y)*ft+epy*fp
              ez=e0z+(etz-e0z)*ft+epz*fp

              esq=(ex*ex+ey*ey+ez*ez)
              sx=ey*bz-by*ez
              sy=ez*bx-bz*ex
              sz=ex*by-bx*ey

              txa=tx-sx
              tya=ty-sy
              tza=tz-sz
              eea=ee-0.5*(esq+bsq)

              if(ieos .eq. 0) then
                h=1.+gam*te/(gam-1.)
                dh=gam/(gam-1.)
              elseif(ieos .eq. 1) then
                h=(5./2.)*te+(3./2.)*sqrt(te**2+(4./9.))
                dh=(5./2.)+(3./2.)*te/sqrt(te**2+(4./9.))
              elseif(ieos .eq. 3) then
                h=(12.*te**2+8.*te+2.)/(3.*te+2.)
                dh=4.-6./(3.*te**2+2.)**2
              endif

              ssq=txa**2+tya**2+tza**2
              gfsq=1.+ssq/(h*dro)**2
              gf=sqrt(gfsq)
              dgf=-dh*(gfsq-1.)/(h*gf)

              f=h*gf-(te/gf)-(eea/dro)
              tmp1=(te/h)*(gfsq-1.)/gfsq
!              df=(dh/gf)*(1.-tmp1)-(1./dgf)              
              df=gf*dh+h*dgf-(1./gf)+te*dgf

              delta=f/df

              xxn=te-delta              

              if(xxn .lt. 0.d0) then
                irecov=2
                write(6,*) ' >> te  < 0 in recov3' &
                          ,' at i, j, k:',i, j, k
                write(6,*) ' >> te', xxn &
                          ,', at x1, x2, x3:',x1(i),x2(j),x3(k)
              endif

              te=xxn

              if(ieos .eq. 0) then
                h=1.+gam*te/(gam-1.)
              elseif(ieos .eq. 1) then
                h=(5./2.)*te+(3./2.)*sqrt(te**2+(4./9.))
              elseif(ieos .eq. 3) then
                h=(12.*te**2+8.*te+2.)/(3.*te+2.)
              endif

              gfsq=1.+ssq/(h*dro)**2
              gf=sqrt(gfsq)
              ro=dro/gf
              pp=dro*te/gf

              vx=txa/(dro*h*gf)
              vy=tya/(dro*h*gf)
              vz=tza/(dro*h*gf)        

            endif
          endif

        enddo

!--- After Newton-Rapson routine ---
! 
        if(irecov .eq. 0) then
          
          te=xxn

          if(ieos .eq. 0) then
            h=1.+gam*te/(gam-1.)
          elseif(ieos .eq. 1) then
            h=(5./2.)*te+(3./2.)*sqrt(te**2+(4./9.))
          elseif(ieos .eq. 3) then
            h=(12.*te**2+8.*te+2.)/(3.*te+2.)
          endif

          gfsq=1.+ssq/(h*dro)**2
          gf=sqrt(gfsq)
          ro=dro/gf
          pp=dro*te/gf

          vx=txa/(dro*h*gf)
          vy=tya/(dro*h*gf)
          vz=tza/(dro*h*gf)        
          vsq=vx**2+vy**2+vz**2
          if(vsq .gt. 1.d0) then 
            write(6,*) 'vt > 1.0' 
            vsq=1.0-dv
            gf=1.0/sqrt(1.0-vsq)
            ro=dro/gf 
            vx=txa/(dro*h*gf)
            vy=tya/(dro*h*gf)
            vz=tza/(dro*h*gf)  
          endif

        endif

!--- Set Primitive variables ---

        if(irecov .eq. 0) then

          uri(1,i,j,k)=ro
          uri(2,i,j,k)=vx
          uri(3,i,j,k)=vy
          uri(4,i,j,k)=vz
          uri(5,i,j,k)=pp
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)
          uri(16,i,j,k)=sg
        else

          uri(1,i,j,k)=ro1
          uri(2,i,j,k)=vx1
          uri(3,i,j,k)=vy1
          uri(4,i,j,k)=vz1
          uri(5,i,j,k)=pp1
          uri(6,i,j,k)=uu(6,i,j,k)/dro
          uri(7,i,j,k)=bx
          uri(8,i,j,k)=by
          uri(9,i,j,k)=bz
          uri(10,i,j,k)=ex
          uri(11,i,j,k)=ey
          uri(12,i,j,k)=ez
          uri(13,i,j,k)=uu(13,i,j,k)
          uri(14,i,j,k)=uu(14,i,j,k)
          uri(15,i,j,k)=uu(15,i,j,k)          
          uri(16,i,j,k)=sg1
        endif

      enddo
    enddo
  enddo
!
!=====================================================================@
!
  return
end subroutine recov6a
!
